#include "sim.h"

int writeReg(int value, int reg)
{
 if (reg == 0)  /* *** Trap?  Ignore for now *** */;
 else regs[reg]= value;
}

#define MSB_MASK  0x80000000
#define OVERFLOW(s1,s2,s3) ((((s1) & MSB_MASK) == ((s2) & MSB_MASK)) ? (((s3) & MSB_MASK) != ((s1) & MSB_MASK)) : 0)

void do_unimp()
{
 printf("Unimplemented instruction\n");
 trapNo= UNIMP_TRAP;
}

void do_add()
{
 int old_rs1= regs[rs1], old_rs2= regs[rs2];
#ifdef TRACE
 printf("ADD r%d, r%d, r%d\n", rd, rs1, rs2);
#endif
 writeReg(regs[rs1]+regs[rs2], rd);
 if (rd != 0 && OVERFLOW(old_rs1, old_rs2, regs[rd]))
  trapNo= OVERFLOW_TRAP;
}

void do_addi()
{
 int old_rs1= regs[rs1];
#ifdef TRACE
 printf("ADDI r%d, r%d, #%d\n", rd, rs1, extra);
#endif
 writeReg(regs[rs1]+extra, rd);
 if (rd != 0 && OVERFLOW(old_rs1, extra, regs[rd]))
  trapNo= OVERFLOW_TRAP;
}

void do_addu()
{
#ifdef TRACE
 printf("ADDU r%d, r%d, r%d\n", rd, rs1, rs2);
#endif
 writeReg(regs[rs1]+regs[rs2], rd);
}

void do_addui()
{
#ifdef TRACE
 printf("ADDUI r%d, r%d, #%d\n", rd, rs1, extra);
#endif
 writeReg(regs[rs1] + (extra & 0xffff), rd);
 /* We and extra with fff, as we sign-extended it from 16 bits when we parsd
    it, and now we're un-sign-extending it, as we want it to be unsigned
    after all. */
}

void do_and()
{
#ifdef TRACE
 printf("AND r%d, r%d, r%d\n", rd, rs1, rs2);
#endif
 writeReg(regs[rs1] & regs[rs2], rd);
}

void do_andi()
{
#ifdef TRACE
 printf("ANDI r%d, r%d, #%d\n", rd, rs1, extra);
#endif
 writeReg(regs[rs1] & (extra & 0xffff), rd);
}

/*
 * BRANCH & JUMP ROUTINES:
 * If these cause the pc to become misaligned, it will be picked up
 * at the next fetch and returned as a TRAP_MISALIGNED
 *  *** This could be more friendly & check before pc is set, but that's less
 *  *** like a real machine. ***
 */

void do_beqz()
{
#ifdef TRACE
 printf("BEQZ r%d, %d\n", rs1, extra);
#endif
 if (regs[rs1] == 0) pc += extra;
}

void do_bfpf()
{
#ifdef TRACE
 printf("BFPF %d\n", extra);
#endif
 trapNo= FP_TRAP;
}

void do_bfpt()
{
#ifdef TRACE
 printf("BFPT %d\n", extra);
#endif
 trapNo= FP_TRAP;
}

void do_bnez()
{
#ifdef TRACE
 printf("BNEZ r%d, %d\n", rs1, extra);
#endif
 if (regs[rs1])
  pc += extra;
}

void do_j()
{
#ifdef TRACE
 printf("J %d\n", extra);
#endif
 pc += extra;
}

void do_jal()
{
#ifdef TRACE
 printf("JAL %d\n", extra);
#endif
 regs[31]= pc;  /* Which is already pointing to the next instruction */
 pc += extra;
}

void do_jalr()
{
#ifdef TRACE
 printf("JALR r%d\n", rs1);
#endif
 regs[31]= pc;
 pc= regs[rs1];
}

void do_jr()
{
#ifdef TRACE
 printf("JR r%d\n", rs1);
#endif
 pc= regs[rs1];
}

void do_lb()
{
 int tmp, value;
#ifdef TRACE
 printf("LB %d(r%d), r%d\n", extra, rs1, rd);
#endif
 if (!RMem(&value, regs[rs1]+extra, 1))  /* Read byte */
 {
  if (value & 0x80) value += 0xffffff00;  /* Sign extend */
  else value &= 0xff;
  writeReg(value, rd);
 }
}

void do_lbu()
{
 int tmp, value;
#ifdef TRACE
 printf("LBU %d(r%d), r%d\n", extra, rs1, rd);
#endif
 if (!RMem(&value, regs[rs1]+extra, 1))  /* Read byte */
 {
  value &= 0xff;
  writeReg(value, rd);
 }
}

void do_ld()
{
#ifdef TRACE
 printf("LD %d(r%d), r%d\n", extra, rs1, rd);
#endif
 trapNo= FP_TRAP;
}

void do_lf()
{
#ifdef TRACE
 printf("LF %d(r%d), r%d\n", extra, rs1, rd);
#endif
 trapNo= FP_TRAP;
}

void do_lh()
{
 int tmp, value;
#ifdef TRACE
 printf("LH %d(r%d), r%d\n", extra, rs1, rd);
#endif
 if (!RMem(&value, regs[rs1]+extra, 2))  /* Read short */
 {
  if (value & 0x8000) value += 0xffff0000;  /* Sign extend */
  else value &= 0xffff;
  writeReg(value, rd);
 }
}

void do_lhi()
{
#ifdef TRACE
 printf("LHI %d\n", extra);
#endif
 writeReg(extra << 16, rd);  /* Load-hi, loads hi 16 bits into rd */
}

void do_lhu()
{
 int tmp, value;
#ifdef TRACE
 printf("LHU %d(r%d), r%d\n", extra, rs1, rd);
#endif
 if (!RMem(&value, regs[rs1]+extra, 2))  /* Read short */
 {
  value &= 0xffff;
  writeReg(value, rd);
 }
}

void do_lw()
{
 int tmp, value;
#ifdef TRACE
 printf("LW %d(r%d), r%d\n", extra, rs1, rd);
#endif
 if (!RMem(&value, regs[rs1]+extra, 4))
  writeReg(value, rd);
}

void do_movd()
{
#ifdef TRACE
 printf("MOVD \n");
#endif
 trapNo= FP_TRAP;
}

void do_movf()
{
#ifdef TRACE
 printf("MOVF \n");
#endif
 trapNo= FP_TRAP;
}

void do_movfp2i()
{
#ifdef TRACE
 printf("MOVFP2I \n");
#endif
 trapNo= UNIMP_TRAP;
}

void do_movi2fp()
{
#ifdef TRACE
 printf("MOVI2FP \n");
#endif
 trapNo= UNIMP_TRAP;
}

void do_movi2s()
{
#ifdef TRACE
 printf("MOVI2S \n");
#endif
 trapNo= UNIMP_TRAP;;
}

void do_movs2i()
{
#ifdef TRACE
 printf("MOVS2I \n");
#endif
 trapNo= UNIMP_TRAP;
}

void do_or()
{
#ifdef TRACE
 printf("OR r%d, r%d, r%d\n", rd, rs1, rs2);
#endif
 writeReg(regs[rs1] | regs[rs2], rd);
}

void do_ori()
{
#ifdef TRACE
 printf("ORI r%d, r%d, #%d\n", rd, rs1, extra);
#endif
 writeReg(regs[rs1] | (extra & 0xffff), rd);
}

void do_rfe()
{
#ifdef TRACE
 printf("RFE \n");
#endif
 trapNo= UNIMP_TRAP;
}

void do_sb()
{
#ifdef TRACE
 printf("SB %d(r%d), r%d\n", extra, rs1, rd);
#endif
 WMem(regs[rd], regs[rs1]+extra, 1);  /* Write byte */
}

void do_sd()
{
#ifdef TRACE
 printf("SD \n");
#endif
 trapNo= FP_TRAP;
}

void do_seq()
{
#ifdef TRACE
 printf("SEQ r%d, r%d, r%d\n", rd, rs1, rs2);
#endif
 writeReg(regs[rs1] == regs[rs2], rd);
}

void do_seqi()
{
#ifdef TRACE
 printf("SEQI r%d, r%d, #%d\n", rd, rs1, extra);
#endif
 writeReg(regs[rs1] == extra, rd);
}

void do_sequ()
{
#ifdef TRACE
 printf("SEQU r%d, r%d, r%d\n", rd, rs1, rs2);
#endif
  writeReg((unsigned int)regs[rs1] == (unsigned int)regs[rs2], rd);
}

void do_sequi()
{
#ifdef TRACE
 printf("SEQUI r%d, r%d, #%d\n", rd, rs1, extra);
#endif
 writeReg((unsigned)regs[rs1] == (unsigned)extra, rd);
}

void do_sf()
{
#ifdef TRACE
 printf("SF \n");
#endif
 trapNo= FP_TRAP;
}

void do_sge()
{
#ifdef TRACE
 printf("SGE r%d, r%d, r%d\n", rd, rs1, rs2);
#endif
 writeReg(regs[rs1] >= regs[rs2], rd);
}

void do_sgei()
{
#ifdef TRACE
 printf("SGEI r%d, r%d, #%d\n", rd, rs1, extra);
#endif
 writeReg(regs[rs1] >= extra, rd);
}

void do_sgeu()
{
#ifdef TRACE
 printf("SGEU r%d, r%d, r%d\n", rd, rs1, rs2);
#endif
  writeReg((unsigned int)regs[rs1] >= (unsigned int)regs[rs2], rd);
}

void do_sgeui()
{
#ifdef TRACE
 printf("SGEUI r%d, r%d, #%d\n", rd, rs1, extra);
#endif
 writeReg((unsigned)regs[rs1] >= (unsigned)extra, rd);
}
void do_sgt()
{
#ifdef TRACE
 printf("SGT r%d, r%d, r%d\n", rd, rs1, rs2);
#endif
 writeReg(regs[rs1] > regs[rs2], rd);
}

void do_sgti()
{
#ifdef TRACE
 printf("SGTI r%d, r%d, #%d\n", rd, rs1, extra);
#endif
 writeReg(regs[rs1] > extra, rd);
}

void do_sgtu()
{
#ifdef TRACE
 printf("SGTU r%d, r%d, r%d\n", rd, rs1, rs2);
#endif
  writeReg((unsigned int)regs[rs1] > (unsigned int)regs[rs2], rd);
}

void do_sgtui()
{
#ifdef TRACE
 printf("SGTUI r%d, r%d, #%d\n", rd, rs1, extra);
#endif
 writeReg((unsigned)regs[rs1] > (unsigned)extra, rd);
}

void do_sh()
{
#ifdef TRACE
 printf("SB %d(r%d), r%d\n", extra, rs1, rd);
#endif
 WMem(regs[rd], regs[rs1]+extra, 2);  /* Write short */
}

void do_sle()
{
#ifdef TRACE
 printf("SLE r%d, r%d, r%d\n", rd, rs1, rs2);
#endif
 writeReg(regs[rs1] <= regs[rs2], rd);
}

void do_slei()
{
#ifdef TRACE
 printf("SLEI r%d, r%d, #%d\n", rd, rs1, extra);
#endif
 writeReg(regs[rs1] <= extra, rd);
}

void do_sleu()
{
#ifdef TRACE
 printf("SLEU r%d, r%d, r%d\n", rd, rs1, rs2);
#endif
  writeReg((unsigned int)regs[rs1] <= (unsigned int)regs[rs2], rd);
}

void do_sleui()
{
#ifdef TRACE
 printf("SLEUI r%d, r%d, #%d\n", rd, rs1, extra);
#endif
 writeReg((unsigned)regs[rs1] <= (unsigned)extra, rd);
}

void do_sll()
{
#ifdef TRACE
 printf("SLL r%d, r%d, r%d\n", rd, rs1, rs2);
#endif
 writeReg(regs[rs1] << (regs[rs2] & 0x1f /* 31 shift max.  32 == 0 */), rd);
}

void do_slli()
{
#ifdef TRACE
 printf("SLLI r%d, r%d, #%d\n", rd, rs1, extra & 0x1f);
#endif
 writeReg(regs[rs1] << (extra & 0x1f), rd);
}

void do_slt()
{
#ifdef TRACE
 printf("SLT r%d, r%d, r%d\n", rd, rs1, rs2);
#endif
 writeReg(regs[rs1] < regs[rs2], rd);
}

void do_slti()
{
#ifdef TRACE
 printf("SLTI r%d, r%d, #%d\n", rd, rs1, extra);
#endif
 writeReg(regs[rs1] < extra, rd);
}

void do_sltu()
{
#ifdef TRACE
 printf("SLTU r%d, r%d, r%d\n", rd, rs1, rs2);
#endif
  writeReg((unsigned int)regs[rs1] < (unsigned int)regs[rs2], rd);
}

void do_sltui()
{
#ifdef TRACE
 printf("SLTUI r%d, r%d, #%d\n", rd, rs1, extra);
#endif
 writeReg((unsigned)regs[rs1] < (unsigned)extra, rd);
}

void do_sne()
{
#ifdef TRACE
 printf("SNE r%d, r%d, r%d\n", rd, rs1, rs2);
#endif
 writeReg(regs[rs1] != regs[rs2], rd);
}

void do_snei()
{
#ifdef TRACE
 printf("SNEI r%d, r%d, #%d\n", rd, rs1, extra);
#endif
 writeReg(regs[rs1] != extra, rd);
}

void do_sneu()
{
#ifdef TRACE
 printf("SNEU r%d, r%d, r%d\n", rd, rs1, rs2);
#endif
  writeReg((unsigned int)regs[rs1] != (unsigned int)regs[rs2], rd);
}

void do_sneui()
{
#ifdef TRACE
 printf("SNEUI r%d, r%d, #%d\n", rd, rs1, extra);
#endif
 writeReg((unsigned)regs[rs1] != (unsigned)extra, rd);
}

void do_sra()
{
#ifdef TRACE
 printf("SRA r%d, r%d, r%d\n", rd, rs1, rs2);
#endif
 writeReg(regs[rs1] >> (regs[rs2] & 0x1f /* 31 shift max.  32 == 0 */), rd);
}

void do_srai()
{
#ifdef TRACE
 printf("SRAI r%d, r%d, #%d\n", rs1, extra & 0x1f, rd);
#endif
 writeReg(regs[rs1] >> (extra & 0x1f), rd);
}

void do_srl()
{
 unsigned int tmp;
#ifdef TRACE
 printf("SRL r%d, r%d, r%d\n", rd, rs1, rs2);
#endif
 tmp= regs[rs1];
 tmp >>= (regs[rs2] & 0x1f);
 writeReg(tmp, rd);
}

void do_srli()
{
 unsigned int tmp;
#ifdef TRACE
 printf("SRLI r%d, r%d, #%d\n", rs1, extra & 0x1f, rd);
#endif
 tmp= regs[rs1];
 tmp >>= (extra & 0x1f);
 writeReg(tmp, rd);

}

void do_sub()
{
 int old_rs1= regs[rs1], old_rs2= regs[rs2];
#ifdef TRACE
 printf("SUB r%d, r%d, r%d\n", rd, rs1, rs2);
#endif
 writeReg(regs[rs1]-regs[rs2], rd);
 if (rd != 0 && OVERFLOW(old_rs1, -old_rs2, regs[rd]))
  trapNo= OVERFLOW_TRAP;
}

void do_subi()
{
 int old_rs1= regs[rs1];
#ifdef TRACE
 printf("SUBI r%d, r%d, #%d\n", rd, rs1, extra);
#endif
 writeReg(regs[rs1]-extra, rd);
 if (rd != 0 && OVERFLOW(old_rs1, -extra, regs[rd]))
  trapNo= OVERFLOW_TRAP;
}

void do_subu()
{
#ifdef TRACE
 printf("SUBU r%d, r%d, r%d\n", rd, rs1, rs2);
#endif
 writeReg(regs[rs1]-regs[rs2], rd);
}

void do_subui()
{
#ifdef TRACE
 printf("SUBUI r%d, r%d, #%d\n", rd, rs1, extra);
#endif
 writeReg(regs[rs1] - (extra & 0xffff), rd);
 /* We and extra with fff, as we sign-extended it from 16 bits when we parsd
    it, and now we're un-sign-extending it, as we want it to be unsigned
    after all. */
}

void do_sw()
{
#ifdef TRACE
 printf("SW %d(r%d), r%d\n", extra, rs1, rd);
#endif
 WMem(regs[rd], regs[rs1]+extra, 4);  /* Write word */
}

void do_trap()
{
#ifdef TRACE
 printf("TRAP %d\n", extra & 0x3ffffff);
#endif
 trapNo= extra & 0x3ffffff;  /* Un-sign-extend trap # */
}

void do_xor()
{
#ifdef TRACE
 printf("XOR r%d, r%d, r%d\n", rd, rs1, rs2);
#endif
 writeReg(regs[rs1] ^ regs[rs2], rd);
}

void do_xori()
{
#ifdef TRACE
 printf("XORI r%d, r%d, #%d\n", rd, rs1, extra);
#endif
 writeReg(regs[rs1] ^ extra, rd);
}

void do_nop()
{
#ifdef TRACE
 printf("NOP \n");
#endif
}
